# dagtimers : Nicely displayed timers.
# Copyright (C) 2023  Michael P. Lane
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from time import sleep
from unittest import TestCase
from dagtimers import Timers, _START_TIME_DEFAULT, _INDENT


class TestTimers(TestCase):

    def test_start(self):

        timers = Timers()
        timers.start("hello")
        self.assertIn(
            "hello",
            timers._dag.keys()
        )
        self._assert_node(timers._dag, "hello", False, True, 0.0, [])
        self._assert_path(timers, ["hello"])
        timers.start("hi")
        self._assert_node(
            timers._dag, "hello", False, True, 0.0, [
                (timers._dag["hello"][2], "hi", False, True, 0.0, [])
            ]
        )
        self._assert_path(timers, ["hello", "hi"])
        timers.start("sup")
        self._assert_node(
            timers._dag, "hello", False, True, 0.0, [
                (timers._dag["hello"][2], "hi", False, True, 0.0, [
                    (timers._dag["hello"][2]["hi"][2], "sup", False, True, 0.0, [])
                ])
            ]
        )
        self._assert_path(timers, ["hello", "hi", "sup"])

    def test_stop(self):

        timers = Timers()
        timers.start("hello")
        self._assert_node(timers._dag, "hello", False, True, 0.0, [])
        self._assert_path(timers, ["hello"])
        sleep(0.25)
        timers.stop()
        self._assert_node(timers._dag, "hello", True, False, 0.25, [])
        self._assert_path(timers, [])
        timers.start("hello")
        self._assert_node(timers._dag, "hello", False, False, 0.25, [])
        self._assert_path(timers, ["hello"])
        sleep(0.5)
        timers.stop()
        self._assert_node(timers._dag, "hello", True, False, 0.75, [])
        self._assert_path(timers, [])
        timers.start("hello")
        timers.start("hi")
        self._assert_node(timers._dag, "hello", False, False, 0.75, [
            (timers._dag["hello"][2], "hi", False, True, 0.0, [])
        ])
        self._assert_path(timers, ["hello", "hi"])
        sleep(0.1)
        timers.stop()
        self._assert_node(timers._dag, "hello", False, False, 0.75, [
            (timers._dag["hello"][2], "hi", True, False, 0.1, [])
        ])
        self._assert_path(timers, ["hello"])
        timers.start("hello")
        self._assert_node(timers._dag, "hello", False, False, 0.75, [
            (timers._dag["hello"][2], "hi", True, False, 0.1, []),
            (timers._dag["hello"][2], "hello", False, True, 0.0, [])
        ])
        self._assert_path(timers, ["hello", "hello"])
        sleep(0.2)
        timers.stop()
        self._assert_node(timers._dag, "hello", False, False, 0.75, [
            (timers._dag["hello"][2], "hi", True, False, 0.1, []),
            (timers._dag["hello"][2], "hello", True, False, 0.2, [])
        ])
        self._assert_path(timers, ["hello"])
        timers.stop()
        self._assert_node(timers._dag, "hello", True, False, 1.05, [
            (timers._dag["hello"][2], "hi", True, False, 0.1, []),
            (timers._dag["hello"][2], "hello", True, False, 0.2, [])
        ])
        self._assert_path(timers, [])

    def test_pretty_print(self):

        timers = Timers()
        self._assert_indent_profile(timers.pretty_print(), [0])
        timers.start("hi")
        self._assert_indent_profile(timers.pretty_print(), [0, 0])
        sleep(0.05)
        timers.stop()
        self._assert_indent_profile(timers.pretty_print(), [0, 0])
        timers.start("hi")
        self._assert_indent_profile(timers.pretty_print(), [0, 0])
        sleep(0.2)
        timers.stop()
        self._assert_indent_profile(timers.pretty_print(), [0, 0])
        timers.start("startyyyyy")
        self._assert_indent_profile(timers.pretty_print(), [0, 0, 0])
        sleep(0.1)
        timers.stop()
        self._assert_indent_profile(timers.pretty_print(), [0, 0, 0])
        timers.start("hi")
        sleep(0.01)
        self._assert_indent_profile(timers.pretty_print(), [0, 0, 0])
        timers.start("hello")
        sleep(0.01)
        self._assert_indent_profile(timers.pretty_print(), [0, 0, 0, 1])
        timers.stop()
        self._assert_indent_profile(timers.pretty_print(), [0, 0, 0, 1])
        timers.stop()
        self._assert_indent_profile(timers.pretty_print(), [0, 0, 1, 1, 0])



    def _assert_node(self, node, timer_name, default_start_time, zero_elasped, elapsed_lb, children):

        self.assertIn(
            timer_name,
            node.keys()
        )

        if default_start_time:
            self.assertEqual(
                node[timer_name][0],
                _START_TIME_DEFAULT
            )

        else:

            try:
                self.assertNotEqual(
                    node[timer_name][0],
                    _START_TIME_DEFAULT
                )

            except AssertionError:
                raise

        if zero_elasped:
            self.assertEqual(
                node[timer_name][1],
                0
            )

        else:

            try:
                self.assertNotEqual(
                    node[timer_name][1],
                    0
                )

            except AssertionError:
                raise

            self.assertGreaterEqual(
                node[timer_name][1],
                elapsed_lb
            )

        for child_args in children:
            self._assert_node(*child_args)

        self.assertEqual(
            len(node[timer_name][2]),
            len(children)
        )

    def _assert_path(self, timers, exp_path_names):

        self.assertEqual(
            timers._path[0],
            timers._dag
        )
        self.assertEqual(
            len(timers._path),
            len(exp_path_names) + 1
        )
        node = timers._dag

        for exp_path_name, path_node in zip(exp_path_names, timers._path[1:]):

            node = node[exp_path_name][2]
            self.assertEqual(
                node,
                path_node
            )

        self.assertEqual(
            timers._parents,
            exp_path_names
        )

    def _assert_indent_profile(self, pretty_print, profile):

        try:
            self.assertEqual(
                len(profile),
                pretty_print.count("\n")
            )

        except AssertionError:
            raise

        for level, line in zip(profile, pretty_print[:-1].split("\n")):

            self.assertEqual(
                line[ : level * len(_INDENT)],
                _INDENT * level
            )
            self.assertNotEqual(
                line[level * len(_INDENT) : (level + 1) * len(_INDENT)],
                _INDENT
            )










# dagtimers : Nicely displayed timers.
# Copyright (C) 2023  Michael P. Lane
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import math
from contextlib import contextmanager
from time import time

__all__ = ["Timers", "TimerError"]

_START_TIME_DEFAULT = -1
_UNACCOUNTED_STR = "<Unaccounted>"
_EXCESS_STR = "<Excess>"
_INDENT = "    |"
_START_TIME_INDEX = 0
_ELAPSED_INDEX = 1
_CHILDREN_INDEX = 2
_NUM_STARTS_INDEX = 3

class TimerError(RuntimeError): pass

class Timers:

    def __init__(self):

        self._starts = {}
        self._num_start = 0
        self._dag = {}
        self._excess = 0
        self._path = [self._dag]
        self._parents = []
        self._last_pretty_print_time = _START_TIME_DEFAULT

    def time_iter(self, timer_name, iter_):

        started = False

        try:

            self.start(timer_name)
            started = True

            for item in iter_:

                self.stop()
                started = False
                yield item
                self.start(timer_name)
                started = True

            self.stop()
            started = False

        finally:

            if started:
                self.stop()

    @contextmanager
    def time_cm(self, timer_name, cm):

        started = False

        try:

            self.start(timer_name)
            started = True

            with cm as yield_:

                self.stop()
                started = False
                yield yield_
                self.start(timer_name)
                started = True

            self.stop()
            started = False

        finally:

            if started:
                self.stop()

    @contextmanager
    def time(self, timer_name):

        try:
            self.start(timer_name)
            yield

        finally:
            self.stop()

    def start(self, timer_name):

        excess_start_time = time()
        self._num_start += 1

        try:
            _, elapsed, next_node_in_path, num_starts = self._path[-1][timer_name]

        except KeyError:

            next_node_in_path = {}
            self._path[-1][timer_name] = (time(), 0, next_node_in_path, 1)

        else:
            self._path[-1][timer_name] = (time(), elapsed, next_node_in_path, num_starts + 1)

        self._path.append(next_node_in_path)
        self._parents.append(timer_name)
        self._excess += time() - excess_start_time

    def stop(self):

        excess_start_time = time()

        if len(self._parents) == 0:
            raise TimerError("No timers currently running.")

        timer_name = self._parents[-1]
        del self._path[-1]
        start_time, elapsed, dag, num_starts = self._path[-1][timer_name]
        self._path[-1][timer_name] = (_START_TIME_DEFAULT, elapsed + time() - start_time, dag, num_starts)
        self._excess += time() - excess_start_time
        del self._parents[-1]

    def _pretty_print(self, indent, root, parent_timer_name, parent):

        ret = ""

        if len(root) == 0:
            return ret

        has_parent = parent_timer_name is not None
        ordered = sorted(
            [timer_name for timer_name in root.keys() if root[timer_name][_START_TIME_INDEX] == _START_TIME_DEFAULT],
            key = lambda timer_name: -root[timer_name][_ELAPSED_INDEX]
        )
        ordered.extend(
            [timer_name for timer_name in root.keys() if root[timer_name][_START_TIME_INDEX] != _START_TIME_DEFAULT]
        )
        accounted = sum(t[_ELAPSED_INDEX] for t in root.values())
        max_num_chars = max(len(timer_name) for timer_name in root.keys())

        if has_parent:

            max_num_chars = max(max_num_chars, len(_UNACCOUNTED_STR))
            parent_running = parent[parent_timer_name][0] != _START_TIME_DEFAULT

            if not parent_running:

                parent_elapsed = parent[parent_timer_name][_ELAPSED_INDEX]
                unaccounted = parent_elapsed - accounted

                if parent_elapsed != 0:
                    prop = (unaccounted / parent_elapsed) * 100

                else:
                    prop = 100

                if parent_elapsed != 0:
                    ret += _INDENT * indent + f"<Unaccounted> : {unaccounted:.7f} ({prop:.3f}%)\n"

        else:

            max_num_chars = max(max_num_chars, len(_EXCESS_STR))
            parent_elapsed = accounted
            parent_running = False

        for timer_name in ordered:

            elapsed = root[timer_name][_ELAPSED_INDEX]
            running = root[timer_name][_START_TIME_INDEX] != _START_TIME_DEFAULT
            ret += _INDENT * indent + f"{timer_name: <{max_num_chars}}"

            if not running:

                ret += f" : {elapsed:.7f}"

                if has_parent and not parent_running:

                    if parent_elapsed != 0:
                        prop = (elapsed / parent_elapsed) * 100

                    else:
                        prop = 100

                    ret += f" ({prop:.3f}%)"

                if self._excess != 0 and elapsed != 0:

                    inside_log = self._num_start * elapsed / (self._excess * root[timer_name][_NUM_STARTS_INDEX])
                    ret += f" [{math.floor(math.log10(inside_log))}]"

                else:
                    ret += f" [inf]"

                ret += "\n"

            else:
                ret += " : RUNNING\n"

            if len(root[timer_name][2]) > 0:
                ret += self._pretty_print(indent + 1, root[timer_name][2], timer_name, root)

        return ret

    def pretty_print(self):

        ret = self._pretty_print(0, self._dag, None, None)

        if len(self._dag) > 0:
            max_num_chars = max(len(timer_name) for timer_name in self._dag.keys())

        else:
            max_num_chars = 0

        max_num_chars = max(max_num_chars, len(_EXCESS_STR))
        return f"{_EXCESS_STR: <{max_num_chars}} : {self._excess:.7f}\n" + ret



# dagtimers : Nicely displayed timers.
# Copyright (C) 2023  Michael P. Lane
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from setuptools import setup

setup(
    name = 'dagtimers',
    version = "0.1.2",
    description = "Nicely displayed timers.",
    long_description = "Nicely displayed timers.",
    long_description_content_type="text/plain",
    author = "Michael P. Lane",
    author_email = "mlanetheta@gmail.com",
    url = "https://github.com/automorphis/dagtimers",
    package_dir = {"": "lib"},
    packages = [
        "dagtimers"
    ],
    test_suite = "tests",
    zip_safe=False
)